function plugin({ types: t }) {

    let setupNodes = [];
    let ifNodes = [];

    /* MyRuleVisitor is the top-level visitor.
     */
    const MyRuleVisitor = {
	FunctionDeclaration(path) {
	    const params = path.node.params;
	    let user, context, callback;
	    if (params[0] && params[0].name) {
		user = params[0].name;
	    }
	    if (params[1] && params[1].name) {
		context = params[1].name;
	    }
	    if (params[2] && params[2].name) {
		callback = params[2].name;
	    }
	    const fnParams = { user, context, callback };
	    console.log(fnParams);
	    path.traverse(BlockVisitor, fnParams);
	}
    };

    const BlockVisitor = {
	BlockStatement(path) {
	    path.traverse(MainProcessor, this);
	}
    };

    // TODO: Use babel-types instead
    function objectContains(obj, key, val) {
	if (typeof obj === 'object' &&
	    obj.hasOwnProperty(key) &&
	    obj[key] === val)
	{
	    return obj;
	}
	for (let curKey in obj) {
	    if (typeof obj[curKey] === 'object' &&
		objectContains(obj[curKey], key, val))
	    {
		return obj[curKey];
	    }
	}
	return false;
    }

    function isClientIfStatement(node) {
	let parentObject = objectContains(node.test, 'type', 'MemberExpression');
	if (node.test && parentObject)
	{
	console.log(parentObject.type);
	    console.log('Found it!');
	    // console.log(parentObject);
	    // console.log(parentObject.object.name);
	    // console.log(parentObject.property.name);
	    return true;
	}
	return false;
    }

    /* Checks if something is "default" code, i.e. it does _nothing_ to the rule
     * Only one way to be default code. It should be of the form:
     * callback(null, user, context);
     * If the code is different _at all_, some code is being run, which means
     * it's non-default code.
     */
    function isDefaultCode(path, params) {
	const callExpression = objectContains(path.node, 'type', 'CallExpression');
	if (callExpression && callExpression.callee &&
	    callExpression.callee.name === params.callback)
	{
	    console.log('--- CallExp', callExpression.callee.name);
	}
    }

    function removeCruft(path) {
	if (t.isReturnStatement(path)) {

	}
    }

    let levelSemaphore = 0;
    let currentPart = 0;
    /* This visitor is the "main processor" of generic Statements.
     * Refer to thoughts.txt to understand what "parts" are.
     */
    const MainProcessor = {
	Statement: {
	    enter(path) {
		levelSemaphore++;
		switch (currentPart) {
		case 0:
		    
		    const appliesName = path.scope.generateUidIdentifier('applies');
		    const appliesFalse = t.variableDeclaration(
			'let',
			[t.variableDeclarator(appliesName, t.booleanLiteral(false))]
		    );
		    path.insertBefore(appliesFalse);
		    currentPart = 1;
		    break;
		case 1:
		    if (path.node.type === 'IfStatement' &&
			isClientIfStatement(path.node))
		    {
			currentPart = 2;
			ifNodes.push(path.node);
			// console.log('Pushed');
			break;
		    }
		    
		    // We're extracting setup code only if part 1
		    // isn't done and if we're actually at the function's
		    // root level
		    if (levelSemaphore === 1) {
			console.log('Enter', path.node.type, levelSemaphore);
			setupNodes.push(path.node);
		    }
		    break;
		case 2:
		    /*
		    if (path.node.type === 'ExpressionStatement' ||
			path.node.type === 'ReturnStatement') {
			isDefaultCode(path, this);
		    }*/
		    console.log('Part 2', path.node.type, levelSemaphore);
		    break;
		}
	    },
	    exit(path) {
		levelSemaphore--;
		switch (currentPart) {
		case 1:
		    break;
		}
	    }
	}
    }

    return {
	visitor: MyRuleVisitor
    };
}

module.exports = plugin;
